;; Generated package description from yasnippet.el
(define-package "yasnippet" "0.9.1" "Yet another snippet extension for Emacs." 'nil :url "http://github.com/capitaomorte/yasnippet" :keywords '("convenience" "emulation"))
